## Note: cythonizing this module has less than a 10% impact on
## performance from testing

INVALID_ELEMENT_MSG = "'UintSet' elements must be integers >= 0"
INVALID_ITER_ARG_MSG = "expected UintSet or iterable argument"
NOT_IN_SET_MSG = "element not in UintSet"
POP_EMPTY_MSG = "pop from an empty set"


def count_ones(bigint):
    count = 0
    while bigint:
        count += bigint & 1
        bigint >>= 1
    return count


def get_bit(bigint, index):
    return bool(bigint & (1 << index))


def set_bit(bigint, index):
    return bigint | (1 << index)


def unset_bit(bigint, index):
    if get_bit(bigint, index):
        return bigint ^ (1 << index)
    return bigint


def find_ones(bigint):
    index = 0
    while True:
        if bigint & 1:
            yield index
        bigint >>= 1
        if bigint == 0:
            break
        index += 1


class UintSet:
    def __init__(self, elements=None, bits=0):
        self._bits = bits
        if elements:
            for e in elements:
                self.add(e)

    def __getitem__(self, index):
        return get_bit(self._bits, index)

    def __len__(self):
        return count_ones(self._bits)

    def copy(self):
        return UintSet(bits=self._bits)

    def add(self, elem):
        try:
            self._bits = set_bit(self._bits, elem)
        except TypeError:
            raise TypeError(INVALID_ELEMENT_MSG)
        except ValueError:
            raise ValueError(INVALID_ELEMENT_MSG)

    def __contains__(self, elem):
        try:
            return get_bit(self._bits, elem)
        except TypeError:
            raise TypeError(INVALID_ELEMENT_MSG)
        except ValueError:
            raise ValueError(INVALID_ELEMENT_MSG)

    def __iter__(self):
        return find_ones(self._bits)

    def __repr__(self):
        elements = ", ".join(str(e) for e in self)
        if elements:
            elements = "{" + elements + "}"
        return f"UintSet({elements})"

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self._bits == other._bits

    def __or__(self, other):
        cls = self.__class__
        if isinstance(other, cls):
            res = cls()
            res._bits = self._bits | other._bits
            return res
        return NotImplemented

    def union(self, *others):
        cls = self.__class__
        res = cls()
        res._bits = self._bits
        for other in others:
            if isinstance(other, cls):
                res._bits |= other._bits
            else:
                try:
                    second = cls(other)
                except TypeError:
                    raise TypeError(INVALID_ITER_ARG_MSG)
                else:
                    res._bits |= second._bits
        return res

    def __and__(self, other):
        cls = self.__class__
        if isinstance(other, cls):
            res = cls()
            res._bits = self._bits & other._bits
            return res
        return NotImplemented

    def intersection(self, *others):
        cls = self.__class__
        res = cls()
        res._bits = self._bits
        for other in others:
            if isinstance(other, cls):
                res._bits &= other._bits
            try:
                second = cls(other)
            except TypeError:
                raise TypeError(INVALID_ITER_ARG_MSG)
            else:
                res._bits &= second._bits
        return res

    def __xor__(self, other):
        cls = self.__class__
        if isinstance(other, cls):
            res = cls()
            res._bits = self._bits ^ other._bits
            return res
        return NotImplemented

    def symmetric_difference(self, other):
        return self ^ other

    def __sub__(self, other):
        cls = self.__class__
        if isinstance(other, cls):
            res = self.copy()
            res._bits ^= self._bits & other._bits
            return res
        return NotImplemented

    def difference(self, *others):
        cls = self.__class__
        res = cls()
        res._bits = self._bits
        for other in others:
            if isinstance(other, cls):
                res._bits ^= self._bits & other._bits
            else:
                try:
                    second = cls(other)
                except TypeError:
                    raise TypeError(INVALID_ITER_ARG_MSG)
                else:
                    res._bits ^= self._bits & second._bits
        return res

    def remove(self, elem):
        try:
            if elem not in self:
                raise KeyError(elem)
            self._bits = unset_bit(self._bits, elem)
        except TypeError:
            raise TypeError(INVALID_ELEMENT_MSG)
        except ValueError:
            raise ValueError(INVALID_ELEMENT_MSG)

    def pop(self):
        if self._bits == 0:
            raise KeyError(POP_EMPTY_MSG)
        elem = next(iter(self))
        self.remove(elem)
        return elem



import sys
import os


sys.path.insert(1, os.path.join(sys.path[0], "../../.."))

import pyarrow
from opteryx.engine.planner.temporal import extract_temporal_filters
from opteryx.utils.arrow import get_metadata


def test(SQL):

    from mabel.utils import timer
    from opteryx.utils.display import ascii_table
    import opteryx
    from opteryx.storage.adapters import DiskStorage

    conn = opteryx.connect(reader=DiskStorage(), partition_scheme=None)
    cur = conn.cursor()

    with timer.Timer():
        # do this to go over the records
        cur.execute(SQL)
        #cur._results = [pyarrow.concat_tables(cur._results)]
        #print("METADATA:", get_metadata(cur._results))

        #print(pyarrow.concat_tables(cur._results))

        print(ascii_table(cur.fetchmany(size=10), limit=10))
        [a for a in cur.fetchmany(1000)]
        print(json.dumps(cur.stats, indent=2))


if __name__ == "__main__":

    import json
    import sqloxide

    # SQL = "SELECT count(*) from `tests.data.zoned` where followers < 10 group by followers"
    # SQL = "SELECT username, count(*) from `tests.data.tweets` group by username"
    SQL = "SELECT COUNT(user_verified) FROM tests.data.set as ds"

    # SQL = """
    # SELECT DISTINCT user_verified, MIN(followers), MAX(followers), COUNT(*)
    #  FROM tests.data.huge
    # GROUP BY user_verified
    # """

    # SQL = "SELECT username from `tests.data.tweets`"

    SQL = "SELECT * FROM $satellites"
    SQL = "SELECT COUNT(*) FROM $satellites"
    SQL = "SELECT MAX(planetId), MIN(planetId), SUM(gm), count(*) FROM $satellites group by planetId"
    SQL = "SELECT upper(name), length(name) FROM $satellites WHERE magnitude = 5.29"
    SQL = "SELECT planetId, Count(*) FROM $satellites group by planetId having count(*) > 5"
    SQL = "SELECT * FROM $satellites order by magnitude, name"
    SQL = "SELECT AVG(gm), MIN(gm), MAX(gm), FIRST(gm), COUNT(*) FROM $satellites GROUP BY planetId"
    SQL = "SELECT COUNT(name) FROM $satellites;"

    SQL = "SELECT name as what_it_known_as FROM $satellites"
    SQL = "SELECT name, id, planetId FROM $satellites"
    SQL = "SELECT COUNT(*), planetId FROM $satellites GROUP BY planetId"
    SQL = "SELECT ROUND(magnitude) FROM $satellites group by ROUND(magnitude)"
    SQL = "SELECT COUNT(*) FROM $satellites"
    SQL = "SELECT * FROM $satellites WHERE (id = 6 OR id = 7 OR id = 8) OR name = 'Europa'"
    SQL = (
        "SELECT BOOLEAN(planetId) FROM $satellites GROUP BY planetId, BOOLEAN(planetId)"
    )
    SQL = "SELECT planetId as pid, round(magnitude) as minmag FROM $satellites"
    SQL = "SELECT RANDOM() FROM $planets"
    SQL = "SELECT RANDOM() FROM $planets"
    SQL = "SELECT TIME()"
    SQL = "SELECT LOWER('NAME')"
    SQL = "SELECT HASH('NAME')"
    SQL = "SELECT * FROM (VALUES (1,2),(3,4),(340,455)) AS t(a,b)"
    # SQL = "SELECT id - 1, name, mass FROM $planets"
    SQL = "SELECT * FROM tests.data.partitioned WHERE $DATE IN ($$PREVIOUS_MONTH)"
    SQL = """
SELECT * FROM (VALUES 
(7369, 'SMITH', 'CLERK', 7902, '02-MAR-1970', 8000, NULL, 20),
(7499, 'ALLEN', 'SALESMAN', 7698, '20-MAR-1971', 1600, 3000, 30),
(7521, 'WARD', 'SALESMAN', 7698, '07-FEB-1983', 1250, 5000, 30),
(7566, 'JONES', 'MANAGER', 7839, '02-JUN-1961', 2975, 50000, 20),
(7654, 'MARTIN', 'SALESMAN', 7698, '28-FEB-1971', 1250, 14000, 30),
(7698, 'BLAKE', 'MANAGER', 7839, '01-JAN-1988', 2850, 12000, 30),
(7782, 'CLARK', 'MANAGER', 7839, '09-APR-1971', 2450, 13000, 10),
(7788, 'SCOTT', 'ANALYST', 7566, '09-DEC-1982', 3000, 1200, 20),
(7839, 'KING', 'PRESIDENT', NULL, '17-JUL-1971', 5000, 1456, 10),
(7844, 'TURNER', 'SALESMAN', 7698, '08-AUG-1971', 1500, 0, 30),
(7876, 'ADAMS', 'CLERK', 7788, '12-MAR-1973', 1100, 0, 20),
(7900, 'JAMES', 'CLERK', 7698, '03-NOV-1971', 950, 0, 30),
(7902, 'FORD', 'ANALYST', 7566, '04-MAR-1961', 3000, 0, 20),
(7934, 'MILLER', 'CLERK', 7782, '21-JAN-1972', 1300, 0, 10))
AS employees (EMPNO, ENAME, JOB, MGR, HIREDATE, SAL, COMM, DEPTNO);
    """
    SQL = """
    SELECT * FROM tests.data.dated
  FOR DATES BETWEEN '2022-02-03' AND '2022-02-03';
      """
    SQL = """
    SELECT *
      FROM $astronaut, UNNEST(Missions) AS Mission
     WHERE Mission = 'Apollo 8'
    """
    SQL = """
    SELECT Birth_Place['town']
      FROM $astronauts
        """
    SQL = "SELECT Birth_Place['town'] FROM $astronauts WHERE Birth_Place['town'] = 'Warsaw'"
    SQL = (
        "SELECT BOOLEAN(planetId) FROM $satellites GROUP BY planetId, BOOLEAN(planetId)"
    )

    SQL = """
    SELECT name FROM $planets WHERE id IN (SELECT * FROM UNNEST((1,2,3)) as id)
    """
    SQL = "SELECT sum(1) FROM $planets;"
    SQL = "SELECT * FROM table_1 FOR SYSTEM_TIME AS OF '2022-02-02'"
    SQL = "SELECT count(*) as c FROM tests.data.dated as d"
    SQL = "SELECT COUNT(*) FROM $astronauts WHERE $astronauts.a = $astronauts.b"
#    SQL = "SELECT * FROM $satellites CROSS JOIN $astronauts"
    SQL = "SELECT * FROM $satellites INNER JOIN $astronauts USING(id)"
#    SQL = (
#        "SELECT planetId FROM $satellites GROUP BY planetId"
#    )

    _, _, SQL = extract_temporal_filters(SQL)  
    ast = sqloxide.parse_sql(SQL, dialect="mysql")
    print(json.dumps(ast, indent=2))

    print()

    import cProfile

    with cProfile.Profile(subcalls=False) as pr:
        test(SQL)




# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is the base class for nodes in the execution plan.

The initializer accepts a QueryStatistics node which is populated by
different nodes differently to record what happened during the query
execution.
"""
import abc
from typing import Iterable
from opteryx.engine.query_statistics import QueryStatistics


class BasePlanNode(abc.ABC):
    @abc.abstractclassmethod
    def __init__(self, statistics: QueryStatistics, **config):
        pass

    @abc.abstractclassmethod
    def execute(self, data_pages: Iterable) -> Iterable:
        pass

    @property
    def greedy(self):
        return False

    @property
    def name(self):
        return "no name"



import numpy as np
import pyarrow.compute as pc

from opteryx.engine.attribute_types import TOKEN_TYPES
from .helpers import columns_to_array, groupify_array

# Filter functionality
def arr_op_to_idxs(arr, op, value):
    if op in ["=", "=="]:
        return np.where(arr == value)
    elif op in ["!=", "<>"]:
        return np.where(arr != value)
    elif op == "<":
        return np.where(arr < value)
    elif op == ">":
        return np.where(arr > value)
    elif op == "<=":
        return np.where(arr <= value)
    elif op == ">=":
        return np.where(arr >= value)
    elif op == "in":
        # MODIFIED FOR OPTERYX
        # some of the lists are saved as sets, which are faster than searching numpy
        # arrays, even with numpy's native functionality.
        mask = []
        for a in arr:
            mask.append(a in value)
        return np.array(mask)
    elif op == "not in":
        # MODIFIED FOR OPTERYX - see comment above
        mask = []
        for a in arr:
            mask.append(a not in value)
        return np.array(mask)
    elif op == "like":
        return pc.match_like(arr, value)
    elif op == "not like":
        return np.invert(pc.match_like(arr, value))
    elif op == "ilike":
        return pc.match_like(arr, value, ignore_case=True)
    elif op == "not ilike":
        return np.invert(pc.match_like(arr, value, ignore_case=True))
    elif op == "~":
        return pc.match_substring_regex(arr, value)
    else:
        raise Exception("Operand {} is not implemented!".format(op))


def _get_values(table, operand):
    """
    MODIFIED FOR OPTERYX
    This allows us to use two identifiers rather than the original implementation which
    forced <identifier> <op> <literal>
    """
    try:
        if operand[1] == TOKEN_TYPES.IDENTIFIER:
            return table.column(operand[0]).to_numpy()
        else:
            return operand[0]
    except:
        print(table.column_names)


def filters(table, filters):
    filters = [filters] if isinstance(filters, tuple) else filters
    # Filter is a list of (col, op, value) tuples
    idxs = np.arange(table.num_rows)
    for (left_op, op, right_op) in filters:  # =, <>, <, >, <=, >=, in and not in
        # MODIFIED FOR OPTERYX
        f_idxs = arr_op_to_idxs(
            _get_values(table, left_op), op, _get_values(table, right_op)
        )
        idxs = idxs[f_idxs]
    return table.take(idxs)


def ifilters(table, filters):
    # ADDED FOR OPTERYX
    # return the indices so we can do unions (OR) and intersections (AND) on the lists
    # of indices to do complex filters
    filters = [filters] if isinstance(filters, tuple) else filters
    # Filter is a list of (col, op, value) tuples
    idxs = np.arange(table.num_rows)
    for (left_op, op, right_op) in filters:
        f_idxs = arr_op_to_idxs(
            _get_values(table, left_op), op, _get_values(table, right_op)
        )
        idxs = idxs[f_idxs]
    return idxs


# Drop duplicates
def drop_duplicates(table, on=[], keep="first"):
    # Gather columns to arr
    arr = columns_to_array(table, (on if on else table.column_names))

    # Groupify
    dic, counts, sort_idxs, bgn_idxs = groupify_array(arr)

    # Gather idxs
    if keep == "last":
        idxs = (np.array(bgn_idxs) - 1)[1:].tolist() + [len(sort_idxs) - 1]
    elif keep == "first":
        idxs = bgn_idxs
    elif keep == "drop":
        idxs = [i for i, c in zip(bgn_idxs, counts) if c == 1]
    return table.take(sort_idxs[idxs])


# Show for easier printing
def head(table, n=5, max_width=100):
    if table is None:
        print("No data in table")
        return

    if table.num_rows == 0:
        print("No data in table")
        return

    # Extract head data
    t = table.slice(length=n)
    head = {k: list(map(str, v)) for k, v in t.to_pydict().items()}

    # Calculate width
    col_width = list(map(len, head.keys()))
    data_width = [max(map(len, h)) for h in head.values()]

    # Print data
    data = [list(head.keys())] + [
        [head[c][i] for c in head.keys()] for i in range(t.num_rows)
    ]
    for i in range(len(data)):
        adjust = [
            w.ljust(max(cw, dw) + 2)
            for w, cw, dw in zip(data[i], col_width, data_width)
        ]
        print(
            ("Row  " if i == 0 else str(i - 1).ljust(5)) + "".join(adjust)[:max_width]
        )
    print("\n")




from typing import Iterable, List


def fetchmany(pages, limit: int = 1000):
    DEFAULT_CHUNK_SIZE = 1000
    chunk_size = min(limit, DEFAULT_CHUNK_SIZE)
    if chunk_size < 0:
        chunk_size = DEFAULT_CHUNK_SIZE

    def _inner_row_reader():
        for page in pages:
            for batch in page.to_batches(max_chunksize=chunk_size):
                yield from batch.to_pylist()

    index = -1
    for index, row in enumerate(_inner_row_reader()):
        if index == limit:
            return
        yield row

    if index < 0:
        yield {}

def fetchone(pages: Iterable) -> dict:
    return fetchmany(pages=pages, limit=1).pop()


def fetchall(pages) -> List[dict]:
    return fetchmany(pages=pages, limit=-1)


"""
Adapted from:
https://stackoverflow.com/questions/55546027/how-to-assign-arbitrary-metadata-to-pyarrow-table-parquet-columns
"""

def create_table_metadata(column_names, expected_rows, name, aliases):
    retval = {
        "_expected_rows": expected_rows,
        "_name": name,
        "_aliases": [a for a in set(aliases + [name]) if a],
    }
    for column in column_names:
        retval[column] = {"_aliases":[column]}
        for a in retval["_aliases"]:
            retval[column]['_aliases'].append(f"{a}.{column}")
    return retval

def set_metadata(tbl, tbl_meta=None):
    """
    Store table-level metadata as json-encoded byte strings.

    Table-level metadata is stored in the table's schema.

    Args:
        tbl (pyarrow.Table): The table to store metadata in
        tbl_meta: A json-serializable dictionary with table-level metadata.
    """
    import pyarrow as pa
    import orjson
    # Create updated column fields with new metadata
    if tbl_meta:

        fields = []
        for col in tbl.schema.names:
            fields.append(tbl.field(col))

        # Get updated table metadata
        tbl_metadata = tbl.schema.metadata or {}
        for k, v in tbl_meta.items():
            if isinstance(v, bytes):
                tbl_metadata[k] = v
            else:
                tbl_metadata[k] = orjson.dumps(v)

        # Create new schema with updated table metadata
        schema = pa.schema(fields, metadata=tbl_metadata)
        # With updated schema build new table (shouldn't copy data)
        tbl = tbl.cast(schema)

    return tbl

def _decode_metadata(metadata):
    """
    Arrow stores metadata keys and values as bytes. We store "arbitrary" data as
    json-encoded strings (utf-8), which are here decoded into normal dict.
    """
    import orjson

    if not metadata:
        # None or {} are not decoded
        return metadata

    decoded = {}
    for k, v in metadata.items():
        key = k.decode('utf-8')
        val = orjson.loads(v)
        decoded[key] = val
    return decoded


def get_metadata(tbl):
    """Get table metadata as dict."""
    return _decode_metadata(tbl.schema.metadata)



from .pyudorandom import shuffle, items, bin_gcd, indices

__title__ = "pyudorandom"
__version__ = "1.0.0"
__author__ = "Mats Julian Olsen"
__license__ = "MIT"
__all__ = ["shuffle", "items", "bin_gcd", "indices"]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Join Node

This is a SQL Query Execution Plan Node.

This performs a JOIN
"""

import sys
import os

sys.path.insert(1, os.path.join(sys.path[0], "../../../.."))

import io
import numpy as np
import orjson
import pyarrow
from typing import Iterable
from opteryx.utils.arrow import fetchall, get_metadata
from opteryx.engine.query_statistics import QueryStatistics
from opteryx.engine.planner.operations.base_plan_node import BasePlanNode
from opteryx.third_party import pyarrow_ops

JOIN_TYPES = ('CrossJoin', 'INNER', 'LEFT') 

JSON_TYPES = {np.bool_: bool, np.int64: int, np.float64: float}


def _serializer(obj):
    return JSON_TYPES[type(obj)](obj)

def _cross_join(left, right):
    """
    cross joins are the cartesian product of two tables, that is, the returned dataset
    is a full copy of the right dataset for every row in the left dataset - resulting
    in len(left) x len(right) number of records. It's the simplest to do.

    For all joins, we assume the right table is in memory and can be reiterated as
    needed so we're going to step the left table. 
    """
    
    pylist = []

    # if there's collisions, rename the fields
    # start with a naive rename
    left_metadata = None
    right_metadata = get_metadata(right) or {}

    right_columns = right.column_names
    right_columns = [f"{right_metadata.get('_name', 'r')}.{name}" for name in right_columns]
    right = right.rename_columns(right_columns)

    for page in left:

        #if left_metadata is None:
        #    left_metadata = get_metadata(left)

        for batch in page.to_batches(max_chunksize=100):
            for left_row in batch.to_pylist():
                for right_row in fetchall([right]):
                    new_row = {**left_row, **right_row}
                    if len(pylist) > 5000:
                        table = pyarrow.Table.from_pylist(pylist)
                        #table = arrow.set_metadata(table, metadata)
                        yield table
                        pylist = []

                    pylist.append(new_row)

    if len(pylist) > 1000:
        table = pyarrow.Table.from_pylist(pylist)
        #table = arrow.set_metadata(table, metadata)
        yield table



class JoinNode(BasePlanNode):
    def __init__(self, statistics: QueryStatistics, **config):
        self._right_table = config.get("right_table")
        self._join_type = config.get("join_type", "CrossJoin")
        self._on = config.get("join_on")

    @property
    def name(self):
        return f"({self._join_type}) Join"

    def __repr__(self):
        return ""

    def execute(self, data_pages: Iterable) -> Iterable:
        
        if not isinstance(self._right_table, pyarrow.Table):
            self._right_table = pyarrow.concat_tables(self._right_table.execute(None))  # type:ignore

        if self._join_type == "CrossJoin":
          yield from _cross_join(data_pages, self._right_table)

        elif self._join_type == "InnerUsing":
            for page in data_pages:
                yield pyarrow_ops.join(self._right_table, page, self._on)


if __name__ == "__main__":

    import sys
    import os

    sys.path.insert(1, os.path.join(sys.path[0], "../../../../.."))

    from opteryx import samples
    from opteryx.third_party import pyarrow_ops
    from opteryx.utils.display import ascii_table
    from opteryx.utils.arrow import fetchmany

    planets = samples.astronauts()
    satellites = samples.satellites()

#    print(planets.column_names)
#    print(planets.to_string(preview_cols=10))
#    planets.rename_columns(['id', 'planet_name', 'mass', 'diameter', 'density', 'gravity', 'escapeVelocity', 'rotationPeriod', 'lengthOfDay', 'distanceFromSun', 'perihelion', 'aphelion', 'orbitalPeriod', 'orbitalVelocity', 'orbitalInclination', 'orbitalEccentricity', 'obliquityToOrbit', 'meanTemperature', 'surfacePressure', 'numberOfMoons'])

    joint = _cross_join([planets], satellites)

    #right_columns = planets.column_names
    #right_columns = [f"planets.{name}" for name in right_columns]
    #planets = planets.rename_columns(right_columns)

    from opteryx.third_party.pyarrow_ops import join

    print('---')

#    joint = join(planets, satellites, on=["id"])

#    print(joint.to_string())

    print(ascii_table(fetchmany(joint, limit=10), limit=10))

    #print(joint.column_names)


from fastapi import FastAPI, Request
from fastapi.responses import ORJSONResponse
import os
import uvicorn

trino = FastAPI()


@trino.post("/v1/statement", response_class=ORJSONResponse)
async def read_main(request: Request):
    print(request)
    return {
        "id": 1,
        "stats": {"byes_read": 2000},
        "infoUri": "",
        "columns": [
            {"name": "fruit", "type": "varchar"},
            {"name": "color", "type": "varchar"},
        ],
        "data": [
            ["Apple", "green"],
            ["Banana", "yellow"],
            ["Orange", "orange"],
            ["Strawberry", "red"],
        ],
    }


# tell the server to start
if __name__ == "__main__":
    uvicorn.run(
        "trino_api:trino",
        host="0.0.0.0",
        port=int(os.environ.get("PORT", 8080)),
        lifespan="on",
    )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Projection Node

This is a SQL Query Execution Plan Node.

This Node eliminates columns that are not needed in a Relation. This is also the Node
that performs column renames.
"""
from typing import Iterable
from opteryx.engine.attribute_types import TOKEN_TYPES
from opteryx.engine.query_statistics import QueryStatistics
from opteryx.engine.planner.operations.base_plan_node import BasePlanNode
from opteryx.exceptions import SqlError


def replace_wildcards(arg):
    if arg[1] == TOKEN_TYPES.WILDCARD:
        return "*"
    return str(arg[0])


class ProjectionNode(BasePlanNode):
    def __init__(self, statistics: QueryStatistics, **config):
        """
        Attribute Projection, remove unwanted columns and performs column renames.
        """
        self._projection = {}

        projection = config.get("projection", {"*": "*"})
        # print("projection:", projection)
        for attribute in projection:
            if "aggregate" in attribute:
                self._projection[
                    f"{attribute['aggregate']}({','.join([replace_wildcards(a) for a in attribute['args']])})"
                ] = attribute["alias"]

            elif "function" in attribute:
                args = [
                    ((f"({','.join(a[0])})", None) if isinstance(a[0], list) else a)
                    for a in attribute["args"]
                ]
                self._projection[
                    f"{attribute['function']}({','.join([replace_wildcards(a) for a in args])})"
                ] = attribute["alias"]

            elif "identifier" in attribute:
                self._projection[attribute["identifier"]] = attribute["alias"]
            else:
                self._projection[attribute] = None

    def __repr__(self):
        return str(self._projection)

    @property
    def name(self):
        return "Projection"

    def execute(self, data_pages: Iterable) -> Iterable:

        from opteryx.utils import arrow

        # if we have nothing to do, move along
        if self._projection == {"*": None}:
            # print(f"projector yielding *")
            yield from data_pages
            return

        # we can't do much with this until we have a chunk to read the metadata from
        metadata = None

        for page in data_pages:

            # first time round we're going to set the metadata
            if metadata is None:

                original_metadata = arrow.get_metadata(page)

                # build all the aliases, then elinimate collisions
                # in the event that the collision hits an active column name, rename it to the shortest alias

                # where we have an explicit alias that no longer exists, replace it with the shortest alias
                # that ends with the explicit alias




            # we elminimate attributes we don't want
            try:
                projection = []
                existing_columns = page.column_names
                for k, v in self._projection.items():
                    if k in existing_columns:
                        projection.append(k)
                    elif v in existing_columns:
                        projection.append(v)
                page = page.select(projection)  # type:ignore
            except KeyError as e:
                field = str(e).split('"')[1]
                raise SqlError(
                    f"Column not found `{field}` - the column may not exist, or need to be added to the GROUP BY clause. ({', '.join(page.column_names)})"
                )

            # then we rename the attributes
            if any([v is not None for k, v in self._projection.items()]):  # type:ignore
                names = []
                existing_columns = page.column_names
                for k, v in self._projection.items():
                    if v and v in existing_columns:
                        names.append(v)
                    elif k in existing_columns:
                        names.append(k)
                page = page.rename_columns(names)

            yield page



import pyarrow as pa
import json
import numpy as np


def str_to_table(arr):
    arr = arr.to_numpy()
    arr = np.vectorize(json.loads)(arr)
    return pa.Table.from_pydict({k: [dic.get(k, None) for dic in arr] for k in arr[0]})



import re
import string
from functools import lru_cache

VALID_CHARACTERS = string.ascii_letters + string.digits + string.whitespace
REGEX_CHARACTERS = {ch: "\\" + ch for ch in ".^$*+?{}[]|()\\"}


def tokenize(text):
    text = text.lower()
    text = "".join([c for c in text if c in VALID_CHARACTERS])
    return text.split()


def sanitize(text, safe_characters: str = VALID_CHARACTERS):
    return "".join([c for c in text if c in safe_characters])


def wrap_text(text, line_len):
    from textwrap import fill

    def _inner(text):
        for line in text.splitlines():
            yield fill(line, line_len)

    return "\n".join(list(_inner(text)))


# https://codereview.stackexchange.com/a/248421


@lru_cache(4)
def _sql_like_fragment_to_regex(fragment):
    """
    Allows us to accepts LIKE statements to search data
    """
    # https://codereview.stackexchange.com/a/36864/229677
    safe_fragment = "".join([REGEX_CHARACTERS.get(ch, ch) for ch in fragment])
    return re.compile("^" + safe_fragment.replace("%", ".*?").replace("_", ".") + "$")


def like(x, y):
    return _sql_like_fragment_to_regex(y.lower()).match(str(x).lower())


def not_like(x, y):
    return not like(x, y)


def similar_to(x, y):
    return re.compile(y).search(x) != None


def levenshtein_distance(word1, word2):
    """
    https://en.wikipedia.org/wiki/Levenshtein_distance
    :param word1:
    :param word2:
    :return:
    """
    word2 = word2.lower()
    word1 = word1.lower()
    matrix = [[0 for x in range(len(word2) + 1)] for x in range(len(word1) + 1)]

    for x in range(len(word1) + 1):
        matrix[x][0] = x
    for y in range(len(word2) + 1):
        matrix[0][y] = y

    for x in range(1, len(word1) + 1):
        for y in range(1, len(word2) + 1):
            if word1[x - 1] == word2[y - 1]:
                matrix[x][y] = min(
                    matrix[x - 1][y] + 1, matrix[x - 1][y - 1], matrix[x][y - 1] + 1
                )
            else:
                matrix[x][y] = min(
                    matrix[x - 1][y] + 1, matrix[x - 1][y - 1] + 1, matrix[x][y - 1] + 1
                )

    return matrix[len(word1)][len(word2)]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This in a REPL to interactively test the Query Parser, Planner and Optimizer.
"""
import sys
import os

sys.path.insert(1, os.path.join(sys.path[0], "../../.."))

import sqloxide
from opteryx.engine import planner

statement = None


sqloxide.parse_sql(sql="SELECT * FROM t WHERE apple LIKE pear", dialect="ansi")


while 1:

    statement = input("Query: ")

    if statement.lower().startswith("quit"):
        break

    syntax_tree = sqloxide.parse_sql(sql=statement, dialect="ansi")

    print(syntax_tree)



import sys
import os

sys.path.insert(1, os.path.join(sys.path[0], "../../../.."))

from typing import Optional
from opteryx.utils import paths
from opteryx.exceptions import MissingDependencyError
from opteryx.storage import BaseStorageAdapter

try:
    from google.auth.credentials import AnonymousCredentials
    from google.cloud import storage

    google_cloud_storage_installed = True
except ImportError:  # pragma: no cover
    google_cloud_storage_installed = False


class GcsStorage(BaseStorageAdapter):
    def __init__(self, project: Optional[str] = None, credentials=None, **kwargs):
        if not google_cloud_storage_installed:  # pragma: no cover
            raise MissingDependencyError(
                "`google-cloud-storage` is missing, please install or include in requirements.txt"
            )

        #        super().__init__(**kwargs)
        self.project = project
        self.credentials = credentials

    def read_blob(self, blob_name):
        import io

        bucket, object_path, name, extension = paths.get_parts(blob_name)
        bucket = bucket.replace("_", "-")

        blob = get_blob(
            project=self.project,
            bucket=bucket,
            blob_name=object_path + name + extension,
        )
        stream = blob.download_as_bytes()
        return io.BytesIO(stream)

    def get_blob_list(self, partition):
        bucket, object_path, name, extension = paths.get_parts(partition)
        bucket = bucket.replace("_", "-")

        print(bucket, object_path, name, extension)

        # this means we're not actually going to GCP
        if os.environ.get("STORAGE_EMULATOR_HOST") is not None:
            client = storage.Client(
                credentials=AnonymousCredentials(),
                project=self.project,
            )
        else:  # pragma: no cover
            client = storage.Client(project=self.project)

        gcs_bucket = client.get_bucket(bucket)
        blobs = list(client.list_blobs(bucket_or_name=gcs_bucket, prefix=object_path))

        yield from [
            bucket + "/" + blob.name for blob in blobs if not blob.name.endswith("/")
        ]


def get_blob(project: str, bucket: str, blob_name: str):

    # this means we're not actually going to GCP
    if os.environ.get("STORAGE_EMULATOR_HOST") is not None:
        client = storage.Client(
            credentials=AnonymousCredentials(),
            project=project,
        )
    else:  # pragma: no cover
        client = storage.Client(project=project)

    gcs_bucket = client.get_bucket(bucket)
    blob = gcs_bucket.get_blob(blob_name)
    return blob


if __name__ == "__main__":

    ds = GcsStorage()

    ds.get_partitions(
        dataset="test/data/partitioned", start_date="2000-01-01", end_date="2000-01-05"
    )



"""
This compensates for missing temporal table support in the SQL parser.

For information on temporal tables see: 
https://blog.devgenius.io/a-query-in-time-introduction-to-sql-server-temporal-tables-145ddb1355d9

This supports the following syntaxes

- FOR TODAY
- FOR YESTERDAY
- FOR DATE <timestamp>
- FOR DATES BETWEEN <timestamp> AND <timestamp>

"""
import re
from opteryx.utils import dates
import datetime

SQL_PARTS = [
    r"SELECT",
    r"FROM",
    r"FOR",
    r"WHERE",
    r"GROUP BY",
    r"HAVING",
    r"ORDER BY",
    r"LIMIT",
    r"OFFSET",
]


def clean_statement(string):
    """
    Remove carriage returns and all whitespace to single spaces
    """
    _RE_COMBINE_WHITESPACE = re.compile(r"\s+")
    return _RE_COMBINE_WHITESPACE.sub(" ", string).strip().upper()


def sql_parts(string):
    """
    Split a SQL statement into clauses
    """
    reg = re.compile(
        r"(\(|\)|,|;|"
        + r"|".join([r"\b" + i.replace(r" ", r"\s") + r"\b" for i in SQL_PARTS])
        + r")",
        re.IGNORECASE,
    )
    parts = reg.split(string)
    return [part.strip() for part in parts if part.strip() != ""]


def remove_comments(string):
    """
    Remove comments from the string
    """
    pattern = r"(\".*?\"|\'.*?\')|(/\*.*?\*/|--[^\r\n]*$)"
    # first group captures quoted strings (double or single)
    # second group captures comments (//single-line or /* multi-line */)
    regex = re.compile(pattern, re.MULTILINE | re.DOTALL)

    def _replacer(match):
        # if the 2nd group (capturing comments) is not None,
        # it means we have captured a non-quoted (real) comment string.
        if match.group(2) is not None:
            return ""  # so we will return empty to remove the comment
        else:  # otherwise, we will return the 1st group
            return match.group(1)  # captured quoted-string

    return regex.sub(_replacer, string)


def parse_range(range):
    range = range.upper()
    TODAY = datetime.date.today()

    if range == "PREVIOUS_MONTH":
        pass
    if range == "THIS_MONTH":
        pass
    if range == "MONTH(YEAR, MONTH)":
        pass
    if range == "PREVIOUS_CYCLE(ROLL_OVER)":
        pass
    if range == "CYCLE(YEAR, MONTH, ROLL_OVER)":
        pass


def parse_date(date):

    date = date.upper()
    TODAY = datetime.date.today()

    if date == "TODAY":
        return TODAY
    if date == "YESTERDAY":
        return TODAY - datetime.timedelta(days=1)

    parsed_date = dates.parse_iso(date[1:-1])
    if parsed_date:
        return parsed_date.date()


def extract_temporal_filters(sql):

    # prep the statement, by normalizing it
    clean_sql = remove_comments(sql)
    clean_sql = clean_statement(clean_sql)
    parts = sql_parts(clean_sql)

    TODAY = datetime.date.today()
    clearing_regex = None
    start_date = TODAY
    end_date = TODAY

    try:
        pos = parts.index("FOR")
        for_date_string = parts[pos + 1]
        for_date = parse_date(for_date_string)

        if for_date:
            start_date = for_date
            end_date = for_date
            clearing_regex = (
                r"(\bFOR[\n\r\s]+" + for_date_string.replace("'", r"\'") + r"(?!\S))"
            )
        elif for_date_string.startswith("DATES BETWEEN "):
            parts = for_date_string.split(" ")
            start_date = parse_date(parts[2])
            end_date = parse_date(parts[4])
            clearing_regex = (
                r"(FOR[\n\r\s]+DATES[\n\r\s]+BETWEEN[\n\r\s]+"
                + parts[2]
                + r"[\n\r\s]+AND[\n\r\s]+"
                + parts[4]
                + r"(?!\S))"
            )
        elif for_date_string.startswith("DATES IN "):
            raise NotImplementedError("FOR DATES IN not implemented")
            # PREVIOUS MONTH
            # PREVIOUS CYCLE
            # THIS MONTH
            # THIS CYCLE

        if clearing_regex:
            regex = re.compile(clearing_regex, re.MULTILINE | re.DOTALL | re.IGNORECASE)
            sql = regex.sub("\n-- FOR STATEMENT REMOVED\n", sql)

        # swap the order if we need to
        if start_date > end_date:
            start_date, end_date = end_date, start_date
    except:
        pass

    return start_date, end_date, sql



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Dataset Reader Node

This is a SQL Query Execution Plan Node.

This Node reads and parses the data from a dataset into a Table.

We plan to do the following:

USE THE ZONEMAP:
- Respond to simple aggregations using the zonemap, such as COUNT(*)
- Use BRIN and selections to filter out blobs from being read that don't contain
  records which can match the selections.

PARALLELIZE READING:
- As one blob is read, the next is immediately cached for reading
"""
import time
import datetime
from enum import Enum
from typing import Iterable

import orjson
from opteryx.exceptions import DatabaseError
from opteryx.engine.planner.operations import BasePlanNode
from opteryx.engine import QueryStatistics
from opteryx.storage import file_decoders
from opteryx.storage.adapters import DiskStorage
from opteryx.storage.schemes import MabelPartitionScheme


class EXTENSION_TYPE(str, Enum):
    # labels for the file extentions
    DATA = "DATA"
    CONTROL = "CONTROL"

do_nothing = lambda x, y: x

KNOWN_EXTENSIONS = {
    "complete": (do_nothing, EXTENSION_TYPE.CONTROL),
    "ignore": (do_nothing, EXTENSION_TYPE.CONTROL),
    "arrow": (file_decoders.arrow_decoder, EXTENSION_TYPE.DATA),  # feather
    "jsonl": (file_decoders.jsonl_decoder, EXTENSION_TYPE.DATA),
    "orc": (file_decoders.orc_decoder, EXTENSION_TYPE.DATA),
    "parquet": (file_decoders.parquet_decoder, EXTENSION_TYPE.DATA),
    "zstd": (file_decoders.zstd_decoder, EXTENSION_TYPE.DATA),  # jsonl/zstd
}


def _get_sample_dataset(dataset):
    # we do this like this so the datasets are not loaded into memory unless
    # they are going to be used
    from opteryx import samples

    SAMPLE_DATASETS = {
        "$satellites/": samples.satellites(),
        "$planets/": samples.planets(),
        "$astronauts/": samples.astronauts(),
        "$no_table/": samples.no_table(),
    }
    dataset = dataset.lower()
    if dataset in SAMPLE_DATASETS:
        return SAMPLE_DATASETS[dataset]
    raise DatabaseError(f"Dataset not found `{dataset}`.")

class DatasetReaderNode(BasePlanNode):
    def __init__(self, statistics: QueryStatistics, **config):
        """
        The Dataset Reader Node is responsible for reading the relevant blobs
        and returning a Table/Relation.
        """
        from opteryx.engine.planner.planner import QueryPlanner

        TODAY = datetime.date.today()

        self._statistics = statistics
        self._alias, self._dataset = config.get("dataset", [None, None])

        if isinstance(self._dataset, (bytearray, QueryPlanner)):
            return

        self._dataset = self._dataset.replace(".", "/") + "/"
        self._reader = config.get("reader", DiskStorage())
        self._partition_scheme = config.get("partition_scheme", MabelPartitionScheme())

        self._start_date = config.get("start_date", TODAY)
        self._end_date = config.get("end_date", TODAY)

        # pushed down projection
        self._projection = config.get("projection")
        # pushed down selection
        self._selection = config.get("selection")

    def __repr__(self):
        if self._alias:
            return f"{self._dataset} => {self._alias}"
        return self._dataset

    @property
    def name(self):
        return "Reader"

    def execute(self, data_pages: Iterable) -> Iterable:

        from opteryx.engine.planner.planner import QueryPlanner
        from opteryx.utils.arrow import set_metadata, create_table_metadata

        # literal datasets
        if isinstance(self._dataset, bytearray):
            import io
            import pyarrow.json

            yield pyarrow.json.read_json(io.BytesIO(self._dataset))
            return

        # query plans
        if isinstance(self._dataset, QueryPlanner):
            yield from self._dataset.execute()
            return

        # sample datasets
        if self._dataset[0] == "$":
            yield _get_sample_dataset(self._dataset)
            return

        # datasets from storage
        partitions = self._reader.get_partitions(
            dataset=self._dataset,
            partitioning=self._partition_scheme.partition_format(),
            start_date=self._start_date,
            end_date=self._end_date,
        )

        self._statistics.partitions_found += len(partitions)

        partition_structure:dict = {}
        expected_rows = 0

        # Build the list of blobs we're going to read and collect summary statistics
        # so we can use them for decisions later.
        for partition in partitions:

            partition_structure[partition] = {}
            partition_structure[partition]["blob_list"] = []
            self._statistics.partitions_scanned += 1

            # Get a list of all of the blobs in the partition.
            blob_list = self._reader.get_blob_list(partition)

            # remove folders, that's items ending with '/'
            blob_list = [blob for blob in blob_list if not blob.endswith("/")]

            # Track how many blobs we found
            count_blobs_found = len(blob_list)
            self._statistics.count_blobs_found += count_blobs_found

            # Filter the blob list to just the frame we're interested in
            if self._partition_scheme is not None:
                blob_list = self._partition_scheme.filter_blobs(blob_list)
                self._statistics.count_blobs_ignored_frames += count_blobs_found - len(
                    blob_list
                )

            for blob_name in blob_list:

                # the the blob filename extension
                extension = blob_name.split(".")[-1]

                # find out how to read this blob
                decoder, file_type = KNOWN_EXTENSIONS.get(extension, (None, None))

                if file_type == EXTENSION_TYPE.CONTROL:
                    self._statistics.count_control_blobs_read += 1
                    # read the control blob
                    control_blob = self._reader.read_blob(blob_name)
                    # record the number of bytes we're reading
                    self._statistics.bytes_read_control += control_blob.getbuffer().nbytes
                    try:
                        control_dict = {}
                        control_dict = orjson.loads(control_blob.read())
                        expected_rows += control_dict.get("records", 0)
                    except (ValueError, TypeError):
                        pass
                elif file_type == EXTENSION_TYPE.DATA:
                    partition_structure[partition]["blob_list"].append((blob_name, decoder,))
                else:
                    self._statistics.count_unknown_blob_type_found += 1 

        for partition in partitions:

            # we're reading this partition now
            if len(blob_list) > 0:
                self._statistics.partitions_read += 1

            for blob_name, decoder in partition_structure[partition]["blob_list"]:

                # we're going to open this blob
                self._statistics.count_data_blobs_read += 1

                start_read = time.time_ns()

                # Read the blob from storage, it's just a stream of bytes at this point
                blob_bytes = self._reader.read_blob(blob_name)

                # record the number of bytes we're reading
                self._statistics.bytes_read_data += blob_bytes.getbuffer().nbytes

                # interpret the raw bytes into entries
                pyarrow_blob = decoder(blob_bytes, self._projection)  # type:ignore

                self._statistics.time_data_read += time.time_ns() - start_read

                # we should know the number of entries
                self._statistics.rows_read += pyarrow_blob.num_rows
                self._statistics.bytes_processed_data += pyarrow_blob.nbytes

                # write dataset information to the metadata
                metadata = create_table_metadata(
                    pyarrow_blob.column_names,
                    expected_rows,
                    self._dataset.replace("/", ".")[:-1],
                    [self._alias]
                )
                pyarrow_blob = set_metadata(pyarrow_blob, metadata)

                # yield this blob
                yield pyarrow_blob



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Distinct Node

This is a SQL Query Execution Plan Node.

This Node eliminates duplicate records.
"""
from typing import Iterable
from pyarrow import concat_tables
from opteryx.engine.query_statistics import QueryStatistics
from opteryx.engine.planner.operations import BasePlanNode
from opteryx.third_party.pyarrow_ops import drop_duplicates


class DistinctNode(BasePlanNode):
    def __init__(self, statistics: QueryStatistics, **config):
        self._distinct = config.get("distinct", True)

    def __repr__(self) -> str:
        return ""

    def greedy(self):
        return True

    @property
    def name(self):
        return "Distinction"

    def execute(self, data_pages: Iterable) -> Iterable:
        if self._distinct:
            yield drop_duplicates(concat_tables(data_pages))
        yield from data_pages



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This is a Base class for a caching layer.

It's used by injecting an instantiated object into the Reader. If the object in None
we skip any caching, if it's set up, we use it as an aside cache.
"""
import abc
from typing import Optional


class BaseBufferCache(abc.ABC):
    def __init__(self, **kwargs):
        pass

    @abc.abstractclassmethod
    def get(sef, key: bytes) -> Optional[bytes]:
        """
        Overwrite this method to retrieve a value from the cache, or None if the
        value is not in the cache.
        """
        pass

    @abc.abstractclassmethod
    def set(self, key: bytes, value: bytes):
        """
        Overwrite this method to place a value in the cache.
        """
        pass
